/* 
 * Copyright (c)2007-2008, Dustin Spicuzza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the Dustin Spicuzza may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Dustin Spicuzza ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Dustin Spicuzza BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

using System.Threading;

namespace Panes {
	
	partial class PaneBase {

		private static List<PaneBase> RootPanes = new List<PaneBase>();
		private static List<PaneBase> Panes = new List<PaneBase>();
		private static int saId = 0;
		private static int paneMoving = 0;

		/// <summary>
		/// Pane that is currently being moved
		/// </summary>
		private static PaneBase currentMovingPane = null;

		internal static PaneBase CurrentMovingPane {
			get { return PaneBase.currentMovingPane; }
			set {
				// ensure this is always reset when unassigning this
				if ( PaneBase.currentMovingPane != null )
					PaneBase.currentMovingPane.canDisplayDragEdges = true;

				PaneBase.currentMovingPane = value;

				// ensure this is always set when assigning this
				if ( PaneBase.currentMovingPane != null )
					PaneBase.currentMovingPane.canDisplayDragEdges = false;
			}
		}

		/// <summary>
		/// This is called by the moving floating pane anytime it moves. This will
		/// call (recursively) each PaneContainer to determine which one is in focus
		/// at the moment. 
		/// </summary>
		internal static void UpdateDragWindows() {

			// do this, so it doesn't matter if the mouse position changes, as long as it stays consistent
			Point mousePosition = new Point(MousePosition.X, MousePosition.Y);
			bool tryMouse = true;

			// go backwards
			for (int i = RootPanes.Count - 1; i >= 0; i--){

				PaneBase pb = RootPanes[i];

				if ( !pb.Created || pb.IsDisposed )
					continue;

				// if its a match, then keep traversing, but change the tryMouse parameter to
				// prevent it from making a mouse comparison.. just ensure that its not displaying
				// a window or anything
				tryMouse = pb.UpdateDragWindow(mousePosition, tryMouse);
			}

		}


		/// <summary>
		/// Set to true if there is a PaneBase moving
		/// </summary>
		internal static bool PaneMoving {
			get { return PaneBase.paneMoving == 1 ? true : false; }
			set {

				bool foundMover = false;
				PaneBase tmpPB = null;
				DockStyle tmpPos = DockStyle.None;

				int cmp = value ? 1 : 0;

				// make this atomic -- should this have a lock?
				if ( Interlocked.CompareExchange(ref PaneBase.paneMoving, cmp, cmp) == cmp )
					return;

				PaneBase.paneMoving = cmp;

				// iterate through all panes and hide anything they may have displayed
				foreach (PaneBase pb in Panes) {

					// ignore ones that haven't removed themselves from the list, or ones
					// that simply don't exist yet
					if ( pb.IsDisposed || !pb.Created )
						continue;

					if ( !value ){

						pb.positionControls.HideControls();

						if ( CurrentMovingPane != null && !foundMover ) {
							// determine if there was a selected position from this control
							DockStyle position = pb.positionControls.GetSelectedPosition();

							// add the moving container to the selected container
							if ( position != DockStyle.None ) {

								tmpPos = position;
								tmpPB = pb;
								foundMover = true;
								
							}
						}
					}
				}

				if (foundMover) {
					CurrentMovingPane.IsRoot = false;
					tmpPB.AddPane(CurrentMovingPane, tmpPos);
				}

				// always nullify this, its important
				if (!value) {
					CurrentMovingPane = null;
					PositionControls.ResetSelectedPosition();
				}
			}
		}

	}
}
