/* 
 * Copyright (c)2007-2008, Dustin Spicuzza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the Dustin Spicuzza may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Dustin Spicuzza ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Dustin Spicuzza BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Collections.Generic;

using System.Threading;

namespace Panes {


	/// <summary>
	/// This handles the positioning controls, and removes gobbley-gook from them. Each paneBase
	/// class has this class, which connects with the position control group class and moves
	/// and displays them as necessary. 
	/// </summary>
	internal sealed class PositionControls {

		PaneBase parent;
		bool isPc;

		static PositionControlGroup r_Group = null;
		static PositionControlGroup c_Group = null;

		PositionControls() { }

		public PositionControls(PaneBase parent) {
			this.parent = parent;
			isPc = (parent.GetType() == typeof(PaneContainer));
			if (r_Group == null)
				r_Group = new PositionControlGroup(false);
			if (c_Group == null)
				c_Group = new PositionControlGroup(true);
		}

	
		/// <summary>
		/// Shows the controls that things get dragged into.
		/// </summary>
		public void UpdateDragWindows(Point mousePosition) {

			if (isPc && parent.IsRoot)
				r_Group.UpdateDragWindows(parent, mousePosition);
			else
				c_Group.UpdateDragWindows(parent, mousePosition);
		}

		/// <summary>
		/// Hides the Drag controls immediately
		/// </summary>
		public void HideControls() {
			if (isPc && parent.IsRoot)
				r_Group.HideControls(parent);
			else 
				c_Group.HideControls(parent);
		}

		/// <summary>
		/// Returns the position that was selected, if any
		/// </summary>
		public DockStyle GetSelectedPosition(){

			if (isPc && parent.IsRoot)
				return r_Group.GetSelectedPosition(parent);
		
			return c_Group.GetSelectedPosition(parent);
		}

		/// <summary>
		/// Resets the selected positions
		/// </summary>
		static public void ResetSelectedPosition() {
			c_Group.ResetSelectedPosition();
			r_Group.ResetSelectedPosition();
		}
		

	}

	/// <summary>
	/// Implements the entire control group
	/// </summary>
	class PositionControlGroup {

		// the last parent assigned to the group
		private PaneBase lastParent = null;

		private bool isRoot = false;
		private List<Form> Forms = new List<Form>();

		PositionForm pLeft = null;
		PositionForm pRight = null;
		PositionForm pTop = null;
		PositionForm pBottom = null;
		PositionForm pFill = null;

		bool visible = false;
		
		private PositionControlGroup() { }
		public PositionControlGroup(bool isRoot) {

			this.isRoot = isRoot;

			pLeft = new PositionForm(DockStyle.Left);
			pRight = new PositionForm(DockStyle.Right);
			pTop = new PositionForm(DockStyle.Top);
			pBottom = new PositionForm(DockStyle.Bottom);
			pFill = new PositionForm(DockStyle.Fill);

			Forms.Add(pLeft);
			Forms.Add(pRight);
			Forms.Add(pTop);
			Forms.Add(pBottom);
			Forms.Add(pFill);
		}


		/// <summary>
		/// Shows the controls that things get dragged into.
		/// </summary>
		public void UpdateDragWindows(PaneBase parent, Point mousePosition) {

			// strategy here (and others) to avoid CPU usage:
			// only assign control properties if we haven't done it before
			// because assigning them multiple times is really CPU intensive

			// interesting to note here, that it already is pretty CPU intensive. But,
			// its certainly a lot better than it could be.. heh. 

			if (!parent.Created || parent.IsDisposed)
				return;

			if ( !Interlocked.ReferenceEquals(lastParent,parent)) {

				Interlocked.Exchange(ref lastParent, parent);

				Rectangle pos = parent.RectangleToScreen(parent.ClientRectangle);

				pFill.Top = pLeft.Top = pRight.Top = pos.Top + pos.Height / 2 - pLeft.Height / 2;
				pFill.Left = pTop.Left = pBottom.Left = pos.Left + pos.Width / 2 - pTop.Width / 2;

				if ( !isRoot ) {

					pLeft.Left = pos.Left + pLeft.Width / 2;
					pRight.Left = pos.Left + pos.Width - ( pRight.Width + pRight.Width / 2 );

					pTop.Top = pos.Top + pTop.Height / 2;
					pBottom.Top = pos.Top + pos.Height - ( pBottom.Height + pBottom.Height / 2 );

				} else {

					pLeft.Left = pos.Left + pos.Width / 2 - pLeft.Width - pLeft.Width / 2;
					pRight.Left = pos.Left + pos.Width / 2 + pRight.Width / 2;

					pTop.Top = pos.Top + pos.Height / 2 - pTop.Height - pTop.Height / 2;
					pBottom.Top = pos.Top + pos.Height / 2 + pBottom.Height / 2;

				}
			}

			foreach ( PositionForm pf in Forms ) {

				// generally, this shouldn't happen.. but if it does, life goes on
				if ( pf.IsDisposed )
					continue;

				if ( !pf.Created )
					pf.Show();

				if ( !this.visible ) {

					// set the member to be visible and topmost
					pf.TopMost = true;
					pf.Visible = true;
				}

				pf.UpdateDragWindow(mousePosition);
			}

			this.visible = true;

		}

		/// <summary>
		/// Hides the Drag controls immediately, IF you are the parent
		/// </summary>
		public void HideControls(PaneBase parent) {

			// only matters if we were the last relevant parent
			if (!Interlocked.ReferenceEquals(parent,lastParent))
				return;

			PaneBase p = lastParent;

			if ( visible ) {

				// if we have focus, then focus on our parent. This is essential, it fixes nasty bugs
				if ( pLeft.Focused || pRight.Focused || pTop.Focused || pBottom.Focused || pFill.Focused) {
					
					if (p != null && p.Created && !p.IsDisposed) {
						Form prt = p.FindForm();
						if (prt != null)
							prt.Focus();
					}
				}

				this.visible = pLeft.Visible = pRight.Visible = pTop.Visible = pBottom.Visible = pFill.Visible = false;
			}
		}

		/// <summary>
		/// Returns the position that was selected, if any
		/// </summary>
		public DockStyle GetSelectedPosition(PaneBase parent){
			if (Interlocked.ReferenceEquals(parent,lastParent)) {
				foreach (PositionForm pc in Forms)
					if (pc.IsSelected)
						return pc.Position;
			}
			
			return DockStyle.None;
		}

		/// <summary>
		/// Resets the selected position
		/// </summary>
		public void ResetSelectedPosition() {
			foreach (PositionForm pc in Forms)
				pc.IsSelected = false;

			lastParent = null;
		}
	}
	
}
