/* 
 * Copyright (c)2007-2008, Dustin Spicuzza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the Dustin Spicuzza may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Dustin Spicuzza ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Dustin Spicuzza BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.ComponentModel;

using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Text;

namespace Panes {

	/// <summary>
	/// Replacement for the standard windows forms tab control, since it doesn't
	/// do everything we need it to do.. 
	/// </summary>
	class TabPane : PaneBase {

		List<TabPaneItem> TabItems = new List<TabPaneItem>();
		TabPaneStyle tabStyle = TabPaneStyle.Top;
		Bitmap tabBitmap = null;

		HScrollBar scroll = new HScrollBar();

		Brush backBrush = null;
		Pen pen = SystemPens.ButtonShadow;
		Graphics graphics = null;

		private int selectedIndex = -1;
		private int displayedIndex = -1;

		private int SelectedIndex {
			get { return selectedIndex; }
			set {
				// switch to the selected tab
				if ( selectedIndex >= 0 && selectedIndex < TabItems.Count && TabItems[selectedIndex].Control != null )
					TabItems[selectedIndex].Control.Visible = false;

				displayedIndex = selectedIndex = value;

				if ( selectedIndex >= 0 && selectedIndex < TabItems.Count && TabItems[selectedIndex].Control != null )
					TabItems[selectedIndex].Control.Visible = true;
			}
		}

		public TabPane(TabPaneStyle style) {
			TabStyle = style;
			Initialize();
		}

		public TabPane() {
			Initialize();
		}

		public void Initialize() {

			this.IsRoot = false;
			SetStyle(ControlStyles.UserPaint, true);
			SetStyle(ControlStyles.AllPaintingInWmPaint, true);

			this.MinimumSize = new Size(MinimumSize.Width, Font.Height * 3);
			this.BackColor = SystemColors.Window;

			scroll.Visible = false;
			scroll.Scroll += new ScrollEventHandler(scroll_Scroll);
			Controls.Add(scroll);

		}

		void scroll_Scroll(object sender, ScrollEventArgs e) {

			if ( e.Type != ScrollEventType.EndScroll ) {

				// adjust the displayIndex accordingly

				// if going left
				if ( scroll.Value == 0 || e.OldValue > e.NewValue ) {
					for ( int i = 0; i < TabItems.Count; i++ )
						if ( TabItems[i].isShown ) {
							if ( displayedIndex == ( i == 0 ? i : i - 1 ) )
								return;
							else {
								displayedIndex = (i == 0 ? i : i - 1);
								break;
							}
						}
				} else { // if going right
					for ( int i = TabItems.Count - 1; i >= 0; i-- )
						if ( TabItems[i].isShown ) {
							if ( displayedIndex == ( i == TabItems.Count - 1 ? i : i + 1 ) )
								return;
							else {
								displayedIndex = ( i == TabItems.Count - 1 ? i : i + 1 );
								break;
							}
						}
				}

				RedrawTabs();
			}
		}


		[DefaultValue("SystemColors.Window")]
		[Description("Color of the currently selected tab")]
		public override Color BackColor {
			get {
				return base.BackColor;
			}
			set {
				base.BackColor = value;
				if ( backBrush != null )
					backBrush.Dispose();
				backBrush = new SolidBrush(value);
			}
		}


		[DefaultValue("CTabStyle.Top")]
		public TabPaneStyle TabStyle{
			get { return tabStyle; }
			set { 
				tabStyle = value;
				RecreateAllTextWidth();
				RedrawTabs(); 
			}
		}

		public TabPaneItem SelectedItem {
			get {
				if ( SelectedIndex < 0 )
					return null;
				return TabItems[SelectedIndex];
			}
		}

		/// <summary>
		/// Adds a new tab to the control
		/// </summary>
		/// <param name="c"></param>
		/// <param name="text"></param>
		public void AddTabItem(Control c, string text) {
			TabPaneItem cti = new TabPaneItem(c, text);

			using ( Bitmap b = new Bitmap(1, 1) ) {
				using ( Graphics g = Graphics.FromImage(b) ) {
					RecreateTextWidth(cti, g);
				}
			}

			// add the control
			c.Dock = DockStyle.Fill;
			c.Visible = false;
			Controls.Add(c);

			// add the item
			TabItems.Add(cti);

			// if there is no selected item, make this the selected item
			if ( SelectedIndex < 0 )
				SelectedIndex = 0;

			// redraw the tabs to include this one
			RedrawTabs();

		}

		public override void AddPane(PaneBase p, DockStyle position) {

			if ( position != DockStyle.Fill ) {
				base.AddPane(p, position);
				return;
			}

			if ( p.mySplitter != null )
				p.mySplitter.Dispose();
			p.mySplitter = null;

			Type t = p.GetType();

			if (t == typeof(PaneContainer)){

				// remove all of its children, and add them as tabs
				while ( p.Controls.Count > 0 ) {

					Control c = p.Controls[0];
					p.Controls.Remove(c);

					if ( c.GetType().BaseType == typeof(PaneBase) )
						AddPane(c as PaneBase, DockStyle.Fill);
					else
						c.Dispose();
				}

			} else if (t == typeof(TabPane)) {
					
				// grab those tabs, and add them to this, then delete the object

				TabPaneItem cti;

				while ( ( cti = ( p as TabPane ).RemoveTabItem(0) ) != null )
					AddTabItem(cti.Control, cti.ItemText);

				p.Dispose();

			} else if ( t == typeof(Pane) ) {
				AddTabItem(p, p.Text);

			} else {
				throw new Exception("Was not implemented correctly");
			}
		}


		/// <summary>
		/// Removes a tab from the controls. This is only designed for use with AddPane, and not for
		/// general usage.. implementation isnt particularly good
		/// </summary>
		private TabPaneItem RemoveTabItem(int index) {

			if ( index < TabItems.Count && index >= 0 ) {

				TabPaneItem ret = TabItems[index];

				System.Threading.Interlocked.Exchange(ref removing, 1);
				this.Controls.Remove(ret.Control);
				System.Threading.Interlocked.Exchange(ref removing, 0);

				TabItems.Remove(ret);

				return ret;
			}

			return null;
		}


		/// <summary>
		/// Parses the items and creates the tab image from it. This is a semi-nasty function, but 
		/// it mostly works. 
		/// </summary>
		private void RedrawTabs(){

			if (!this.Created)
				return;

			if (Width < 1 || Height < 1 ) {

				if (tabBitmap != null)
					tabBitmap.Dispose();
				tabBitmap = null;

				this.Padding = new Padding((int)pen.Width);
				return;
			}
		

			// create a bitmap
			Bitmap bmp = new Bitmap(Width, Height);
			Graphics g = Graphics.FromImage(bmp);
			
			// fill bmp with background color, setup padding

			if ( SelectedIndex < 0 ) {

				this.Padding = new Padding((int)pen.Width);
				g.FillRectangle(backBrush, 0, 0, Width, Height);
				g.DrawRectangle(pen, 0, 0, Width - pen.Width, Height - pen.Width);

			} else {

				if ( TabStyle == TabPaneStyle.Top ) {
					g.FillRectangle(SystemBrushes.Control, 0, 0, bmp.Width, Font.Height * 2);
					g.FillRectangle(backBrush, 0, Font.Height * 2, bmp.Width, Height);
					Padding = new Padding((int)pen.Width, Font.Height * 2 + (int)pen.Width, (int)pen.Width, (int)pen.Width);
				} else {
					g.FillRectangle(SystemBrushes.Control, 0, Height - Font.Height * 2, bmp.Width, Font.Height * 2);
					g.FillRectangle(backBrush, pen.Width, pen.Width, bmp.Width, Height - (Font.Height * 2 + pen.Width));
					Padding = new Padding((int)pen.Width, (int)pen.Width, (int)pen.Width, Font.Height * 2);
				}

				// using the selected index, work your way out
				float totalWidth = TabItems[SelectedIndex].Width;
				bool goLeft = true;
				bool goRight = true;
				int startIndex = displayedIndex;

				PointF[] sPts = null;
				float h2 = (float)Font.Height / 2;

				

				for ( int i = 1; i < TabItems.Count; i++ ) {

					// go to the right, verify
					if ( !goRight || displayedIndex + i >= TabItems.Count || TabItems[displayedIndex + i].Width + totalWidth >= Width )
						goRight = false;
					else
						totalWidth += TabItems[displayedIndex + i].Width;

					// go to the left, verify
					if ( !goLeft || displayedIndex - i < 0 || TabItems[displayedIndex - i].Width + totalWidth >= Width )
						goLeft = false;
					else {
						totalWidth += TabItems[displayedIndex - i].Width;
						startIndex = displayedIndex - i;
					}

					if ( !goLeft && !goRight )
						break;
				}

				// reset this, reuse it
				totalWidth = 0;

				bool isScrolling = ( startIndex != 0 );

				// offset
				if ( tabStyle == TabPaneStyle.Bottom )
					g.TranslateTransform(0, Height - Font.Height * 2);


				// now, starting with the start index, draw the tabs from the left
				for ( int i = 0; i < TabItems.Count; i++ ) {

					TabPaneItem cti = TabItems[i];

					// check to see if its not visible
					if ( i < startIndex || totalWidth + cti.Width >= Width ) {

						if ( totalWidth + cti.Width >= Width )
							isScrolling = true;

						cti.isShown = false;
						continue;
					}

					TabItems[i].isShown = true;

					// fill the area
					if ( i == SelectedIndex ) {
						g.FillPolygon(backBrush, cti.drawPoints);

						float f1;
						float f2;
						int x = 0;

						if ( TabStyle == TabPaneStyle.Top ) {
							f1 = Font.Height * 2;
							f2 = Height - pen.Width;
						} else {
							f1 = Height - (Font.Height * 2 + pen.Width);
							f2 = 0;
						}

						// create a set of points to outline the box
						if ( SelectedIndex == 0 ) {
							sPts = new PointF[5];
							sPts[4] = new PointF(cti.Width, f1);

						} else {
							sPts = new PointF[6];

							sPts[x++] = new PointF(totalWidth, f1);
							sPts[5] = new PointF(totalWidth + cti.Width, f1);
						}

						sPts[x++] = new PointF(0, f1);
						sPts[x++] = new PointF(0, f2);
						sPts[x++] = new PointF(Width - pen.Width, f2);
						sPts[x++] = new PointF(Width - pen.Width, f1);
						

					} else
						g.FillPolygon(SystemBrushes.Control, cti.drawPoints);

					// then draw an outline around it
					g.DrawLines(pen, cti.drawPoints);

					// then draw some text
					g.DrawString(cti.ItemText, Font, SystemBrushes.WindowText, h2, h2);

					// translate
					g.TranslateTransform(cti.Width, 0);
					totalWidth += cti.Width;
				}

				g.ResetTransform();

				if (sPts != null)
					g.DrawLines(pen, sPts);
				//else
				//	g.DrawRectangle(pen, );

				// determine the scrolling part
				if ( isScrolling ) {
					if ( !scroll.Visible )
						scroll.Visible = true;

					scroll.Height = Font.Height;
					scroll.Width = Font.Height * 2;
					scroll.Left = Width - ( Font.Height * 5 ) / 2;
					if ( TabStyle == TabPaneStyle.Top )
						scroll.Top = (int)h2;
					else
						scroll.Top = ( Height - Font.Height * 2 ) + (int)h2;

				} else {
					if ( scroll.Visible )
						scroll.Visible = false;
				}
			}

			g.Dispose();

			if (tabBitmap != null)
				tabBitmap.Dispose();

			tabBitmap = bmp;

			graphics.DrawImage(tabBitmap, 0, 0);
		}



		/// <summary>
		/// Reset the text width of all items
		/// </summary>
		public void RecreateAllTextWidth() {

			using ( Bitmap b = new Bitmap(1, 1) ) {
				using ( Graphics g = Graphics.FromImage(b) ) {
					foreach ( TabPaneItem item in TabItems ) {
						RecreateTextWidth(item, g);
					}
				}
			}

			RedrawTabs();
		}

		/// <summary>
		/// Manages the text width of a particular item
		/// </summary>
		private void RecreateTextWidth(TabPaneItem cti, Graphics g) {

			float h = (float)Font.Height;
			float h2 = h / 2;
			float h18 = h / 18;
			float hd = h * 2;

			// create a measurement for the text and a drawing path
			cti.Width = (g.MeasureString(cti.ItemText, Font)).Width;
			
			// calculate the points to be drawn
			cti.drawPoints[0] = new PointF(0, hd);
			cti.drawPoints[1] = new PointF(0, 0);
			cti.drawPoints[2] = new PointF(cti.Width + h2, 0);

			cti.drawPoints[3] = new PointF(cti.Width + h - h18 * 4, h18);
			cti.drawPoints[4] = new PointF(cti.Width + h - h18 * 2, h18 * 2);
			cti.drawPoints[5] = new PointF(cti.Width + h - h18, h18 * 4);
			
			cti.drawPoints[6] = new PointF(cti.Width + h, h2);
			cti.drawPoints[7] = new PointF(cti.Width + h, hd);
			
			// if it is the bottom style, then flip the set of points
			if ( tabStyle == TabPaneStyle.Bottom ) {
				g.ScaleTransform(1, -1);
				g.TranslateTransform(0, - (hd - pen.Width));
				g.TransformPoints(CoordinateSpace.World, CoordinateSpace.Page, cti.drawPoints);
				g.ResetTransform();
			}

			cti.Width += h;
		}

		int removing = 0;


		protected override void OnControlRemoved(ControlEventArgs e) {
			base.OnControlRemoved(e);

			if ( removing == 1 )
				return;

			// this number is 2 because of the item hasn't been removed yet.. 
			if ( TabItems.Count == 2 ) {

				System.Threading.Interlocked.Exchange(ref removing, 1);

				PaneBase pb = Controls[1] as PaneBase;

				// set these properties correctly, since its replacing us
				pb.mySplitter = mySplitter;
				pb.IsRoot = IsRoot;
				pb.Dock = Dock;
				pb.Size = Size;
				pb.Visible = true;

				mySplitter = null;

				int idx = Parent.Controls.IndexOf(this);

				// remove from self, and add to parent
				pb.Parent = null;

				Parent.Controls.Add(pb);
				Parent.Controls.SetChildIndex(pb, idx);

				// remove self from parent and dispose
				Parent = null;
				Dispose();

				return;
			}

			// remove and redraw
			for ( int i = 0; i < TabItems.Count; i++ ) {

				TabPaneItem cti = TabItems[i];

				if ( e.Control == cti.Control ) {
					TabItems.Remove(cti);

					if ( i == SelectedIndex ) {

						// select a new selected item
						if ( i < TabItems.Count )
							SelectedIndex = i;
						else if ( i > 0 )
							SelectedIndex = i - 1;
						else
							SelectedIndex = -1;

					}

					RedrawTabs();
					break;
				}
			}
		}

		protected override void OnFontChanged(EventArgs e) {
			base.OnFontChanged(e);
			this.MinimumSize = new Size(MinimumSize.Width, Font.Height * 3);
			RecreateAllTextWidth();
		}

		protected override void OnPaint(PaintEventArgs e) {
			base.OnPaint(e);

			if (tabBitmap != null)
				e.Graphics.DrawImageUnscaled(tabBitmap, 0, 0);
		}



		protected override void OnBackColorChanged(EventArgs e) {
			base.OnBackColorChanged(e);
			RedrawTabs();
		}

		protected override void OnCreateControl() {
			base.OnCreateControl();
			graphics = this.CreateGraphics();
			RecreateAllTextWidth();
		}

		protected override void OnForeColorChanged(EventArgs e) {
			base.OnForeColorChanged(e);
			RedrawTabs();
		}

		protected override void OnResize(EventArgs e) {
			base.OnResize(e);

			Graphics g = graphics;
			graphics = this.CreateGraphics();
			if ( g != null )
				g.Dispose();
			RedrawTabs();
		}

		protected override void OnMouseClick(MouseEventArgs e) {
			base.OnMouseClick(e);
			if ( Enabled && e.Button == MouseButtons.Left) {

				if ( TabStyle == TabPaneStyle.Top ) {
					if ( e.Y > Font.Height * 2 )
						return;
				} else {
					if ( e.Y < Height - Font.Height * 2 )
						return;
				}

				// figure out which tab was selected
				int total = 0;
				for (int i = 0; i < TabItems.Count; i++){

					if ( !TabItems[i].isShown )
						continue;

					total += (int)TabItems[i].Width;

					if ( e.X < total ) {

						// switch to the selected tab
						SelectedIndex = i;

						RedrawTabs();
						return;
					}
				}
			}
		}


		public class TabPaneItem {

			public Control Control = null;
			public string ItemText = null;

			// per-instance items managed by the tab control
			internal bool isShown = true;
			internal float Width = -1;
			internal PointF[] drawPoints = new PointF[8];

			private TabPaneItem() { }

			public TabPaneItem(Control Control, string ItemText) {
				this.Control = Control;
				this.ItemText = ItemText;
			}

		}

		public enum TabPaneStyle {
			Top,
			Bottom
		}

	}


}
